// kibibu Serpent.h : main header file for the kibibu Serpent DLL
//

#pragma once

#ifndef __AFXWIN_H__
	#error include 'stdafx.h' before including this file for PCH
#endif

#include "resource.h"		// main symbols

#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <string.h>


// comment out the following line to enable float rather than double
// precision
#define SERPENT_DOUBLE_VERSION

// Uncomment the following line to enable Stereo Serpent goodness (mono otherwise)
// #define SERPENT_STEREO

#include "../MDK/mdk.h"
#include "../DSPLIB/dsplib.h"

// Used in File I/O
#define SERPENT_MAGIC_1		((byte)'S')
#define SERPENT_MAGIC_2		((byte)'E')
#define SERPENT_FILE_VERSION 0x01

// enable a high-precision version using doubles
// rather than floats
#ifdef SERPENT_DOUBLE_VERSION

typedef double serp_double;

#else

typedef float serp_double;

#endif

// Enable optimisations
#pragma optimize( "atwpg", on )

// MACHINE SETTINGS
#define miMACHINETYPE MT_GENERATOR

#define miMACHINEFLAGS 0

#ifdef SERPENT_DOUBLE_VERSION

#ifdef SERPENT_STEREO
// hq, stereo
#define miMACHINE_NAME "kibibu Serpent 1.0 (hq,s)"
#define miSHORT_NAME "Serpent (hs)"
#define miMACHINE_AUTHOR "Cameron Foale (kibibu)"
#define miCOMMAND_STRING "Edit..."

#else
// hq, mono
#define miMACHINE_NAME "kibibu Serpent 1.0 (hq,m)"
#define miSHORT_NAME "Serpent (hm)"
#define miMACHINE_AUTHOR "Cameron Foale (kibibu)"
#define miCOMMAND_STRING "Edit..."

#endif

#else

#ifdef SERPENT_STEREO
// lq, stereo
#define miMACHINE_NAME "kibibu Serpent 1.0 (nq,s)"
#define miSHORT_NAME "Serpent (s)"
#define miMACHINE_AUTHOR "Cameron Foale (kibibu)"
#define miCOMMAND_STRING "Edit..."

#else
// lq, mono
#define miMACHINE_NAME "kibibu Serpent 1.0 (nq,m)"
#define miSHORT_NAME "Serpent (m)"
#define miMACHINE_AUTHOR "Cameron Foale (kibibu)"
#define miCOMMAND_STRING "Edit..."

#endif

#endif


#define miMACHINETYPESTRING "[generator]"

#define miABOUTTXT "kibibu serpent"

#ifdef SERPENT_STEREO
#define miISALWAYSSTEREO 1
#else
#define miISALWAYSSTEREO 0
#endif

// Constants relating to sound generated
// MAX_FREQ is NOT the highest frequency, but the highest OSCILLATOR
// frequency. i.e. serpent will generate higher frequencies.
#define MAX_FREQ	(5000)
#define MAX_LENGTH	(2000)
#define MAX_TRACK	(5.0)
#define MIN_TRACK	(-5.0)
#define WORD_MAX	(65534)
#define WORD_NO		(65535)

#define LENGTH_MAX	(5000)
#define LENGTH_INFINITE	(5001)
#define LENGTH_NO	(WORD_NO)


enum WaveType
{
	sine = 0,
	saw,
	square,
	triangle,
	noise1,
	noise2
};



#define WAVETYPE_MAX (Wavetype.periodic_noise)

// Filters. Stolen from Doof 1.5, which was "borrowed" from somewhere else
// From memory, the State Variable Filter on musicdsp.org
#define FILTERTYPE_LOW			0
#define FILTERTYPE_HIGH			1
#define FILTERTYPE_NOTCH		2
#define FILTERTYPE_BAND			3

// Unit Types
#define UNIT_TICKS				0
#define UNIT_TICKS_OVER_128		1
#define UNIT_MS					2
#define UNIT_SAMPLES			3

#define UNIT_MIN				0
#define UNIT_MAX				3
#define UNIT_NO					(0xFF)

// fractional accuracy - for phase etc
#define MAX_FRAC 1000

// Low volume. Used to determine whether muting is done yet
// and whether filter is still generating samples
// For filtering, check all filter outputs (similar to checking history) to 
// ensure oscillations are finished
#define LOW_VOLUME				(0.0001f)

#define AMP_MAX	200


// Global Parameters
CMachineParameter const paraAmp = 
{ pt_word, "Amp","Global Amp",0,AMP_MAX,WORD_NO,0,0 };

// Track Parameters
CMachineParameter const paraNote = 
{ pt_note, "Note","Note",NOTE_MIN,NOTE_OFF,NOTE_NO,0,0 };

CMachineParameter const paraNote2 = 
{ pt_note, "Note2","Note2",NOTE_MIN,NOTE_OFF,NOTE_NO,0,0 };

CMachineParameter const paraSustainLength = 
{ pt_word, "SusLength","Sustain Length",0,LENGTH_INFINITE,LENGTH_NO,MPF_STATE,0x04};

CMachineParameter const paraLengthUnits = 
{ pt_byte, "LengthUnit","Length Unit",UNIT_MIN,UNIT_MAX,UNIT_NO,MPF_STATE,UNIT_TICKS};

CMachineParameter const paraSustainForceLoop = 
{ pt_switch, "SusForceLoop","Finish Sustain Loop",SWITCH_OFF,SWITCH_ON,SWITCH_NO,MPF_STATE,SWITCH_ON};

CMachineParameter const paraSmooth = 
{ pt_switch, "Smooth","Smooth Parameters",SWITCH_OFF,SWITCH_ON,SWITCH_NO,MPF_STATE,SWITCH_OFF};

CMachineParameter const paraCtrl1 = 
{ pt_word, "Ctrl1","Control 1",0,WORD_MAX,WORD_NO,MPF_STATE,WORD_NO};

CMachineParameter const paraCtrl2 = 
{ pt_word, "Ctrl2","Control 2",0,WORD_MAX,WORD_NO,MPF_STATE,WORD_NO};

CMachineParameter const paraCtrl3 = 
{ pt_word, "Ctrl3","Control 3",0,WORD_MAX,WORD_NO,MPF_STATE,WORD_NO};

CMachineParameter const paraCtrl4 = 
{ pt_word, "Ctrl4","Control 4",0,WORD_MAX,WORD_NO,MPF_STATE,WORD_NO};


#define miMAX_TRACKS		12
#define miMIN_TRACKS		1
#define miNUMGLOBALPARAMETERS	1
#define miNUMTRACKPARAMETERS 10
#define miNUMATTRIBUTES		0
#define miVERSION "1.0"



#define SEP_FLAGS_NONE			(0x00)
#define SEP_FLAGS_SUSTAIN_START (0x01)	// both SUSTAIN_START and SUSTAIN_END can be applied to the one point
#define SEP_FLAGS_SUSTAIN_END	(0x02)
#define SEP_FLAGS_CONTROL		(0x04)

#define SEP_FLAGS_SPLINEPOINT	(0x08)

#define SEP_FLAGS_FIXED_VALUE	(0x10)


// A single point on the envelope
class CSerpentEnvelopePoint
{
public:
	serp_double t;	// time
	float v;	// value
	int flags;

	CSerpentEnvelopePoint()
	{
		t = v = 0.0f;
		flags = SEP_FLAGS_NONE;

	}

	inline bool IsControl()
	{
		return ((flags & SEP_FLAGS_CONTROL) > 0);
	}

	inline bool IsPoint()
	{
		return !((flags & SEP_FLAGS_CONTROL) > 0);
	}

	inline bool IsSustainStart()
	{
		return (flags & SEP_FLAGS_SUSTAIN_START) > 0;
	}

	inline bool IsSustainEnd()
	{
		return (flags & SEP_FLAGS_SUSTAIN_END) > 0;
	}

	// straight vector maths
	inline CSerpentEnvelopePoint operator+ (CSerpentEnvelopePoint sep)
	{
		CSerpentEnvelopePoint sepRet;

		sepRet.t = t + sep.t;
		sepRet.v = v + sep.v;

		return sepRet;
	}

	inline CSerpentEnvelopePoint operator/ (float fDiv)
	{
		CSerpentEnvelopePoint sepRet;

		sepRet.t = t / fDiv;
		sepRet.v = v / fDiv;

		return sepRet;
	}

	inline CSerpentEnvelopePoint operator* (serp_double dRatio)
	{
		CSerpentEnvelopePoint sepRet;

		sepRet.t = t * dRatio;
		sepRet.v = v * (float)dRatio;

		return sepRet;
	}

	inline CSerpentEnvelopePoint operator* (CSerpentEnvelopePoint & p)
	{
		CSerpentEnvelopePoint sepRet;

		sepRet.t = t * p.t;
		sepRet.v = v * p.v;

		return sepRet;
	}

		
};


// p2 is only valid if this is the last span, or if
// CSerpentEnvelope::UpdateLastPoints() has been called
class CSerpentEnvelopeSpan
{
public:
	CSerpentEnvelopePoint p1; 
	CSerpentEnvelopePoint c1;
	CSerpentEnvelopePoint c2;

	CSerpentEnvelopePoint p2;

	CSerpentEnvelopeSpan()
	{
		c1.flags = SEP_FLAGS_CONTROL;
		c2.flags = SEP_FLAGS_CONTROL;

		ResetValues();
	}

	inline void ResetValues()
	{
		p1.t = 0.0;
		p1.v = 1.0f;
		p2.t = 1.0;
		p2.v = 0.0f;
		c1.t = 0.25;
		c1.v = 0.5f;
		c2.t = 0.75;
		c2.v = 0.5f;
	}


	// returns true if the points are in the correct order
	inline bool IsValid()
	{
		return((p1.t < c1.t) && (c1.t < c2.t) && (c2.t < p2.t));
	}

	inline void MakeValid()
	{
		if(IsValid()) return;

		if(c1.t < p1.t) c1.t = p1.t;
		if(c2.t > p2.t) c2.t = p2.t;

		if(c2.t < c1.t) c1.t = c2.t = (c1.t + c2.t) / 2;
	}
		
};

// Used for speeding up wave generation
#define LINEAR_ENVELOPE_SEGMENTS 1024
#define ONE_OVER_LINEAR_ENVELOPE_SEGMENTS (1.0/LINEAR_ENVELOPE_SEGMENTS)

class CSerpentEnvelope;

class CSerpentLinearEnvelope
{
public:

	serp_double tInterval;
	serp_double tTotal;
	bool bSustain;

	float fPoints[LINEAR_ENVELOPE_SEGMENTS];

	void PopulateFromSerpentEnvelope(CSerpentEnvelope * pEnv);
	inline float GetValueAtTime(serp_double t);
};


// SerpentEnvelope class
class CSerpentEnvelope
{
public:
	CSerpentEnvelopeSpan env;
	CSerpentEnvelope * pNextEnvelope;
	CSerpentEnvelope * pPrevEnvelope;

	CSerpentLinearEnvelope * pLinearEnv;

	bool bLinked;

	bool bFlat;
	float fFlatValue;

	bool bIsEntirelyZero;

	bool CalcZero()
	{
			
		if((0.0f == env.p1.v) && (0.0f == env.c1.v) && (0.0f == env.c2.v) && (0.0f == env.p2.v))
		{
			if(pNextEnvelope)
			{
				// not the last envelope,
				// but this one is zero,
				// so check the next one
				bIsEntirelyZero = pNextEnvelope->CalcZero();	
				return bIsEntirelyZero;
			} else {
				// last envelope
				// this one is zero, return true
				bIsEntirelyZero = true;
				return true;
			}
		} else {
			// not zero.
			// return false
			bIsEntirelyZero = false;
			return false;
		}
	}

	int SpanCount()
	{

		if(!pNextEnvelope)
		{
			return 1;
		} else {
			return 1 + pNextEnvelope->SpanCount();
		}
	}

	serp_double GetLastTime()
	{

		if(pNextEnvelope)
		{
			return pNextEnvelope->GetLastTime();
		} else {
			return env.p2.t;
		}
	}

	// deletes all allocated envelopes
	// won't delete self
	void Delete()
	{

		if(pNextEnvelope)
		{
			pNextEnvelope->Delete();
			delete pNextEnvelope;
		}
		pNextEnvelope = NULL;
		pPrevEnvelope = NULL;
	}

	virtual ~CSerpentEnvelope()
	{
		if(pLinearEnv) delete(pLinearEnv);
	}

	void CalcFlat()
	{

		if((env.p1.v == env.c1.v) && (env.c2.v == env.p2.v) && (env.c1.v == env.c2.v))
		{
			this->bFlat = true;	
			fFlatValue = env.p1.v;
		} else {
			this->bFlat = false;
		}

		// this one is flat
		if(pNextEnvelope)
		{
			pNextEnvelope->CalcFlat();
		} 
	}

	void UpdateEnvelopeDetails()
	{
		CalcFlat();
		CalcZero();
		UpdateLastPoints();

		if(pLinearEnv)
		{
			pLinearEnv->PopulateFromSerpentEnvelope(this);
		}
	}


	void Reset()
	{
		Delete();
		env.ResetValues();
		UpdateEnvelopeDetails();
	}

	void ResetFlat(float fVal)
	{
		Delete();
		env.ResetValues();
		env.p1.v = env.c1.v = env.c2.v = env.p2.v = fVal;
		UpdateEnvelopeDetails();
	}

	void ResetCurve(float fv1, float fv2)
	{
		Delete();
		env.ResetValues();
		env.p1.v = env.c1.v = fv1;
		env.c2.v = env.p2.v = fv2;
		UpdateEnvelopeDetails();
	}

	CSerpentEnvelopePoint * GetSustainStart()
	{

		if(env.p1.flags & SEP_FLAGS_SUSTAIN_START)
		{
			// check first point
			return &(env.p1);
		} else if(pNextEnvelope)
		{
			// check next envelope
			return pNextEnvelope->GetSustainStart();
		}
		else if(env.p2.flags & SEP_FLAGS_SUSTAIN_START)
		{
			// check second point
			return &(env.p2);
		} else {
			return NULL;
		}
	}

	CSerpentEnvelopePoint * GetSustainEnd()
	{

		if(env.p1.flags & SEP_FLAGS_SUSTAIN_END)
		{
			// check first point
			return &(env.p1);
		} else if(pNextEnvelope)
		{
			// check next envelope
			return pNextEnvelope->GetSustainEnd();
		}
		else if(env.p2.flags & SEP_FLAGS_SUSTAIN_END)
		{
			// check second point
			return &(env.p2);
		} else {
			return NULL;
		}
	}

	

	void UpdateLastPoints()
	{

		if(pNextEnvelope)
		{
			env.p2 = pNextEnvelope->env.p1;
			pNextEnvelope->UpdateLastPoints();
		}
	}

	void ClearFlags( int iFlag)
	{

		env.p1.flags &= ~iFlag;
		env.p2.flags &= ~iFlag;
		env.c1.flags &= ~iFlag;
		env.c2.flags &= ~iFlag;

		if(pNextEnvelope)
		{
			this->pNextEnvelope->ClearFlags(iFlag);
		}
	}
		
	
	bool MoreSpans()
	{
		return (pNextEnvelope != NULL);
	}

	CSerpentEnvelope()
	{
		pNextEnvelope = NULL;
		pPrevEnvelope = NULL;
		bLinked = false;

		this->pLinearEnv = new CSerpentLinearEnvelope();
		pLinearEnv->PopulateFromSerpentEnvelope(this);		
	}

	CSerpentEnvelope(bool bCreateLinear)
	{
		
		pNextEnvelope = NULL;
		pPrevEnvelope = NULL;
		bLinked = false;

		if(bCreateLinear)
		{
			this->pLinearEnv = new CSerpentLinearEnvelope();
			pLinearEnv->PopulateFromSerpentEnvelope(this);
		} else {
			this->pLinearEnv = NULL;
		}


	}

	serp_double SustainWrap(serp_double t)
	{
		// first, find the time of the sustain start
		CSerpentEnvelopePoint * pSusStart = this->GetSustainStart();
		if(!pSusStart) return t;
		CSerpentEnvelopePoint * pSusEnd = this->GetSustainEnd();
		if(!pSusEnd) return t;

		if(pSusStart == pSusEnd) return pSusStart->t;

		if(t > pSusEnd->t)
		{
			t = (serp_double)fmod(t, pSusEnd->t - pSusStart->t) + pSusStart->t;
		}

		return t;
	}

	float GetValueAtTime(serp_double t, CSerpentEnvelope ** ppFoundEnvelope, bool bLock = true)
	{
		if(pLinearEnv) return (pLinearEnv->GetValueAtTime(t));

		return GetValueAtTimeSlow(t, ppFoundEnvelope, bLock);
	}

	float GetValueAtTimeSlow(serp_double t, CSerpentEnvelope ** ppFoundEnvelope, bool bLock = true)
	{

#ifdef DEBUG
		static int iTotalIterations = 0;
		static int iMaxIterations = 0;
		static int iTotalFetches = 0;
#endif

		// repeat the binary style search
		if(t > env.p2.t && pNextEnvelope)
		{
			return pNextEnvelope->GetValueAtTimeSlow(t,ppFoundEnvelope, false);
		} else {
			if(t >= env.p2.t)
			{
				// if t > p2, return p2's value
				*ppFoundEnvelope = this;
				return env.p2.v;
			}

			// otherwise, we are in the right envelope
			*ppFoundEnvelope = this;

			// check if this is flat		
			if(bFlat)
			{
				*ppFoundEnvelope = this;
				return fFlatValue;
			}

			serp_double u = 0.5;
			serp_double dUInc = 0.25;
			serp_double dCloseEnough = (serp_double)(0.002 * 0.002);	// 2 milliseconds
			serp_double dDistance;
			serp_double tFound;
			float vFound;
			
			int iCount = 0;

			serp_double u2, u3, oneminusu, oneminusu2, oneminusu3;

			do
			{
				u2 = u * u;
				u3 = u * u2;
				oneminusu = (serp_double)(1.0 - u);
				oneminusu2 = oneminusu * oneminusu;
				oneminusu3 = oneminusu2 * oneminusu;

				// only care about t right now

				tFound = (oneminusu3 * env.p1.t) +
					(3 * u * oneminusu2 * env.c1.t) +
					(3 * u2 * oneminusu * env.c2.t) + 
					(u3 * env.p2.t );
				

				dDistance = tFound - t;
				dDistance *= dDistance;

				if(dDistance < dCloseEnough)
				{					
					break;
				} else {
					if(t > tFound) u += (dUInc);
					if(t < tFound) u -= (dUInc);
				}

				dUInc *= 0.5;

				iCount++;

#ifdef DEBUG
				iTotalIterations++;	
#endif

			} while(iCount < 50);	// limit iterations

#ifdef DEBUG
			if(iCount > iMaxIterations) iMaxIterations = iCount;
			iTotalFetches++;
#endif

			vFound = (float)((oneminusu3 * env.p1.v) +
					(3 * u * oneminusu2 * env.c1.v) +
					(3 * u2 * oneminusu * env.c2.v) + 
					(u3 * env.p2.v ));

			return vFound;

		}
	}

	// splits this spline into 2
	void Split()
	{

		// allocate the right span
		CSerpentEnvelope *pRight = new CSerpentEnvelope(false);
		CSerpentEnvelopeSpan left;

		// algo from Foley et. al. pg 508
		// They use P1, P2, P3, P4
		// I have used P1, C1, C2, P2
		// from Foley et al:

		// L2 = (P1 + P2) / 2; H = (P2 + P3)/2; L3 = (L2 + H)/2;
		// R3 = (P3 + P4) / 2; R2 = (H + R3)/2; L4 = R1 = (L3 + R2) / 2
		
		CSerpentEnvelopePoint sepH = (env.c1 + env.c2) / 2;

		left.p1 = env.p1;
		left.c1 = (env.p1 + env.c1) / 2;
		left.c2 = (left.c1 + sepH) / 2;
		
		pRight->env.p2 = env.p2;
		pRight->env.c2 = (env.p2 + env.c2) / 2;
		pRight->env.c1 = (pRight->env.c2 + sepH) / 2;

		left.p2 = pRight->env.p1 = (left.c2 + pRight->env.c1) / 2;

		// points set up
		// set up flags etc
		pRight->env.c2.flags = pRight->env.c1.flags = left.c1.flags = left.c2.flags = SEP_FLAGS_CONTROL;
		// pRight->env.p2 is already a copy of env.p2
		// left.p1 is also a copy of env.p1

		// set self up with new Left points
		env.c1 = left.c1;
		env.c2 = left.c2;
		env.p2 = left.p2;

		// insert pRight into the chain
		pRight->pNextEnvelope = this->pNextEnvelope;
		pRight->pPrevEnvelope = this;

		this->pNextEnvelope = pRight;
		
		// this is now left
	}


	// dRatio should be between 0.0 and 1.0
	void Split(serp_double dRatio)
	{
		// allocate the right span
		CSerpentEnvelope *pRight = new CSerpentEnvelope();
		CSerpentEnvelopeSpan left;

		serp_double dOneMinusRatio = (serp_double)(1.0 - dRatio);

		if(! (dRatio > 0.0) ) return;
		if(! (dRatio < 1.0) ) return;

		// algo from Foley et. al. pg 508
		// They use P1, P2, P3, P4
		// I have used P1, C1, C2, P2
		// from Foley et al:

		// L2 = (P1 + P2) / 2; H = (P2 + P3)/2; L3 = (L2 + H)/2;
		// R3 = (P3 + P4) / 2; R2 = (H + R3)/2; L4 = R1 = (L3 + R2) / 2

		// changed to include ratio
		// L2 = (dRatio* P1) + (dOneMinusRatio * P2) etc
		
		CSerpentEnvelopePoint sepH = (env.c1 * dOneMinusRatio) + (env.c2 * dRatio);

		left.p1 = env.p1;
		left.c1 = (env.p1 * dOneMinusRatio) + (env.c1 * dRatio);
		left.c2 = (left.c1* dOneMinusRatio) + (sepH * dRatio);
		
		pRight->env.p2 = env.p2;
		pRight->env.c2 = (env.p2 * dRatio) + (env.c2 * dOneMinusRatio);
		pRight->env.c1 = (pRight->env.c2 * dRatio) + (sepH * dOneMinusRatio);

		left.p2 = pRight->env.p1 = (left.c2  * dOneMinusRatio) + (pRight->env.c1  * dRatio);

		// points set up
		// set up flags etc
		pRight->env.c2.flags = pRight->env.c1.flags = left.c1.flags = left.c2.flags = SEP_FLAGS_CONTROL;
		// pRight->env.p2 is already a copy of env.p2
		// left.p1 is also a copy of env.p1

		// set self up with new Left points
		env.c1 = left.c1;
		env.c2 = left.c2;
		env.p2 = left.p2;

		// insert pRight into the chain
		pRight->pNextEnvelope = this->pNextEnvelope;
		pRight->pPrevEnvelope = this;

		this->pNextEnvelope = pRight;
		
		// this is now left


	}
		
};





// number of samples per wavetable. Oversampled
// must be a power of 2
#define OVERSAMPLING 4

// MUST MUST MUST be a power of two
#define WAVETABLE_SAMPLES (4096 * OVERSAMPLING)

#define WAVETABLE_SIZE_MASK (WAVETABLE_SAMPLES - 1)

// number of sets of samples per wave type
#define SAMPLE_SETS	12
// calculated from 2 ^ 12 = 4096

#define MAX_FILTER_CUTOFF (16000)

#define OSC_FRAC_MASK (0xFF)
#define OSC_FRAC_SHIFT (8)

#define OSC_FRAC(x) (x & OSC_FRAC_MASK)
#define OSC_WHOLE(x) (x >> OSC_FRAC_SHIFT)


// this is used to speed up modulus to wrap phase to table
// and takes into account the OSC_FRAC_MASK
#define WAVETABLE_SIZE_MASK_OSCILLATOR (0x3FFFFF)


class COscillator
{
public:
	int iPhase;				// phase as an int;
	int iPhaseInc;			// phase increment, whole

	COscillator()
	{
		iPhase = iPhaseInc = 0;
	}

	void Reset()
	{
		iPhase = 0;
	}

	inline void CalculatePhase(float fFreq, int iSamplesPerSecond)
	{
		// calculate phase
		float fPhaseInc = fFreq * WAVETABLE_SAMPLES / iSamplesPerSecond;
		iPhaseInc = int(fPhaseInc * OSC_FRAC_MASK);
		if(iPhaseInc < 0) iPhaseInc = 0;
	}

	// returns previous phase
	// can be used directly as a table lookup
	inline int GetPhaseAndIncrement()
	{
		static int iPrevPhase;

		iPrevPhase = iPhase;
		iPhase = (iPhase + iPhaseInc) & WAVETABLE_SIZE_MASK_OSCILLATOR;

		// convert to table lookup and return
		return OSC_WHOLE(iPhase);
	}

	// returns the whole number of samples that the 
	// oscillator increments by
	inline int GetPhaseInc()
	{
		return OSC_WHOLE(iPhaseInc);
	}
		
};



// the following envelopes will be available
// Note1->Note2
// Pitch (Global Multiplier - pre)
// 
// FM Pitch
// FM ->OSC1
// FM ->OSC2
// FM ->OSC3
// 
// OSC1 Pitch (multiplier)
// OSC1 Amp
// OSC2 Pitch (multiplier)
// OSC2 Amp
// OSC3 Pitch (multiplier)
// OSC3 Amp
//
// Clip Level
// 
// Filter 1 Cutoff
// Filter 1 Resonance
// Filter 1 LP Output
// Filter 1 HP Output
// Filter 1 BP Output
//
// Filter 2 Cutoff
// Filter 2 Resonance
// Filter 2 LP Output
// Filter 2 HP Output
// Filter 2 BP Output
//
// Filter 3 Cutoff
// Filter 3 Resonance
// Filter 3 LP Output
// Filter 3 HP Output
// Filter 3 BP Output

// Amp (Multiplier)

class mi;

// Ye Gads! This is a monster!!!!
// There are better ways to store time/value combinations
// than in separate variables. Same with filter values
class CTrack
{
public:
	// filter bits
	// the cutoff and reso
	float fFilt1Cutoff_L, fFilt1Res_L;
	float fFilt1Damp_L;
	float fFilt1Notch_L, fFilt1Low_L, fFilt1High_L, fFilt1Band_L;

	float fFilt2Cutoff_L, fFilt2Res_L;
	float fFilt2Damp_L;
	float fFilt2Notch_L, fFilt2Low_L, fFilt2High_L, fFilt2Band_L;

	float fFilt3Cutoff_L, fFilt3Res_L;
	float fFilt3Damp_L;
	float fFilt3Notch_L, fFilt3Low_L, fFilt3High_L, fFilt3Band_L;

	float fFilt1Cutoff_R, fFilt1Res_R;
	float fFilt1Damp_R;
	float fFilt1Notch_R, fFilt1Low_R, fFilt1High_R, fFilt1Band_R;

	float fFilt2Cutoff_R, fFilt2Res_R;
	float fFilt2Damp_R;
	float fFilt2Notch_R, fFilt2Low_R, fFilt2High_R, fFilt2Band_R;

	float fFilt3Cutoff_R, fFilt3Res_R;
	float fFilt3Damp_R;
	float fFilt3Notch_R, fFilt3Low_R, fFilt3High_R, fFilt3Band_R;

	mi * serpent;

	COscillator oscFM_L;
	COscillator osc1_L;
	COscillator osc2_L;
	COscillator osc3_L;

	COscillator oscFM_R;
	COscillator osc1_R;
	COscillator osc2_R;
	COscillator osc3_R;

	serp_double fNoteFreq;	// note frequency. hz
	serp_double fNote2Freq;	// note frequency. hz

	int iLengthUnits;
	serp_double dLengthInUnits;	// 
	serp_double dSustainLength;	// seconds
	bool bInfiniteSustain;

	bool bSustaining;
	bool bPlaying_L;
	bool bPlaying_R;

	bool bForceLoop;

	bool bUseNote2;

	void ResetTrack()
	{
		iLengthUnits = UNIT_MS;
		dLengthInUnits = 1000;
		dSustainLength = 1000;
		bInfiniteSustain = false;

		bPlaying_L = false;
		bPlaying_R = false;

		bSustaining = false;
		bForceLoop = true;
		bUseNote2 = false;

		bSmoothingNow_L = false;
		bSmoothingNow_R = false;

		fFilt1Cutoff_L = fFilt1Res_L = fFilt1Damp_L = fFilt1Notch_L = fFilt1Low_L
			= fFilt1High_L = fFilt1Band_L = fFilt2Cutoff_L =  fFilt2Res_L = fFilt2Damp_L
			= fFilt2Notch_L = fFilt2Low_L = fFilt2High_L = fFilt2Band_L
			= fFilt3Cutoff_L = fFilt3Res_L = fFilt3Damp_L = fFilt3Notch_L
			= fFilt3Low_L = fFilt3High_L = fFilt3Band_L = 0.0f;

		fFilt1Cutoff_R = fFilt1Res_R = fFilt1Damp_R = fFilt1Notch_R = fFilt1Low_R
			= fFilt1High_R = fFilt1Band_R = fFilt2Cutoff_R =  fFilt2Res_R = fFilt2Damp_R
			= fFilt2Notch_R = fFilt2Low_R = fFilt2High_R = fFilt2Band_R
			= fFilt3Cutoff_R = fFilt3Res_R = fFilt3Damp_R = fFilt3Notch_R
			= fFilt3Low_R = fFilt3High_R = fFilt3Band_R = 0.0f;

	}

	CTrack()
	{
		ResetTrack();
	}

	// times
	serp_double dTimeTotal;

	serp_double dTimeNote1ToNote2_L;	

	serp_double dTimeOctaveShift_L;

	serp_double dTimeFMPitch_L;
	serp_double dTimeFMOsc1_L;
	serp_double dTimeFMOsc2_L;
	serp_double dTimeFMOsc3_L;

	serp_double dTimePitchOsc1_L;
	serp_double dTimeAmpOsc1_L;

	serp_double dTimePitchOsc2_L;
	serp_double dTimeAmpOsc2_L;

	serp_double dTimePitchOsc3_L;
	serp_double dTimeAmpOsc3_L;

	serp_double dTimeClip_L;

	serp_double dTimeFilter1Cut_L;
	serp_double dTimeFilter1Res_L;
	serp_double dTimeFilter1LP_L;
	serp_double dTimeFilter1HP_L;
	serp_double dTimeFilter1BP_L;

	serp_double dTimeFilter2Cut_L;
	serp_double dTimeFilter2Res_L;
	serp_double dTimeFilter2LP_L;
	serp_double dTimeFilter2HP_L;
	serp_double dTimeFilter2BP_L;

	serp_double dTimeFilter3Cut_L;
	serp_double dTimeFilter3Res_L;
	serp_double dTimeFilter3LP_L;
	serp_double dTimeFilter3HP_L;
	serp_double dTimeFilter3BP_L;

	serp_double dTimeAmp_L;

	serp_double dTimeNote1ToNote2_R;	

	serp_double dTimeOctaveShift_R;

	serp_double dTimeFMPitch_R;
	serp_double dTimeFMOsc1_R;
	serp_double dTimeFMOsc2_R;
	serp_double dTimeFMOsc3_R;

	serp_double dTimePitchOsc1_R;
	serp_double dTimeAmpOsc1_R;

	serp_double dTimePitchOsc2_R;
	serp_double dTimeAmpOsc2_R;

	serp_double dTimePitchOsc3_R;
	serp_double dTimeAmpOsc3_R;

	serp_double dTimeClip_R;

	serp_double dTimeFilter1Cut_R;
	serp_double dTimeFilter1Res_R;
	serp_double dTimeFilter1LP_R;
	serp_double dTimeFilter1HP_R;
	serp_double dTimeFilter1BP_R;

	serp_double dTimeFilter2Cut_R;
	serp_double dTimeFilter2Res_R;
	serp_double dTimeFilter2LP_R;
	serp_double dTimeFilter2HP_R;
	serp_double dTimeFilter2BP_R;

	serp_double dTimeFilter3Cut_R;
	serp_double dTimeFilter3Res_R;
	serp_double dTimeFilter3LP_R;
	serp_double dTimeFilter3HP_R;
	serp_double dTimeFilter3BP_R;

	serp_double dTimeAmp_R;

	// end times

	bool bSmooth;
	bool bSmoothingNow_L;
	bool bSmoothingNow_R;

	// values
	float fValueNote1ToNote2_L;	

	float fValueOctaveShift_L;

	float fValueFMPitch_L;
	float fValueFMOsc1_L;
	float fValueFMOsc2_L;
	float fValueFMOsc3_L;

	float fValuePitchOsc1_L;
	float fValueAmpOsc1_L;

	float fValuePitchOsc2_L;
	float fValueAmpOsc2_L;

	float fValuePitchOsc3_L;
	float fValueAmpOsc3_L;

	float fValueClip_L;

	float fValueFilter1Cut_L;
	float fValueFilter1Res_L;
	float fValueFilter1LP_L;
	float fValueFilter1HP_L;
	float fValueFilter1BP_L;

	float fValueFilter2Cut_L;
	float fValueFilter2Res_L;
	float fValueFilter2LP_L;
	float fValueFilter2HP_L;
	float fValueFilter2BP_L;

	float fValueFilter3Cut_L;
	float fValueFilter3Res_L;
	float fValueFilter3LP_L;
	float fValueFilter3HP_L;
	float fValueFilter3BP_L;

	float fValueAmp_L;

	float fValueNote1ToNote2_R;	

	float fValueOctaveShift_R;

	float fValueFMPitch_R;
	float fValueFMOsc1_R;
	float fValueFMOsc2_R;
	float fValueFMOsc3_R;

	float fValuePitchOsc1_R;
	float fValueAmpOsc1_R;

	float fValuePitchOsc2_R;
	float fValueAmpOsc2_R;

	float fValuePitchOsc3_R;
	float fValueAmpOsc3_R;

	float fValueClip_R;

	float fValueFilter1Cut_R;
	float fValueFilter1Res_R;
	float fValueFilter1LP_R;
	float fValueFilter1HP_R;
	float fValueFilter1BP_R;

	float fValueFilter2Cut_R;
	float fValueFilter2Res_R;
	float fValueFilter2LP_R;
	float fValueFilter2HP_R;
	float fValueFilter2BP_R;

	float fValueFilter3Cut_R;
	float fValueFilter3Res_R;
	float fValueFilter3LP_R;
	float fValueFilter3HP_R;
	float fValueFilter3BP_R;

	float fValueAmp_R;
	// end values

	void ResetOscillators(bool bLeft)
	{
		if(bLeft)
		{
			osc1_L.Reset();
			osc2_L.Reset();
			osc3_L.Reset();
			oscFM_L.Reset();
		} else {
			osc1_R.Reset();
			osc2_R.Reset();		
			osc3_R.Reset();
			oscFM_R.Reset();
		}
	}

	void ResetTimes(bool bLeft)
	{
		if(bLeft)
		{
			dTimeNote1ToNote2_L= dTimeOctaveShift_L=
			dTimeFMPitch_L=	dTimeFMOsc1_L=
			dTimeFMOsc2_L= dTimeFMOsc3_L=
			dTimePitchOsc1_L= dTimeAmpOsc1_L=
			dTimePitchOsc2_L= dTimeAmpOsc2_L=
			dTimePitchOsc3_L= dTimeAmpOsc3_L=
			dTimeClip_L= dTimeFilter1Cut_L=
			dTimeFilter1Res_L= dTimeFilter1LP_L=
			dTimeFilter1HP_L= dTimeFilter1BP_L=
			dTimeFilter2Cut_L= dTimeFilter2Res_L=
			dTimeFilter2LP_L= dTimeFilter2HP_L=
			dTimeFilter2BP_L= dTimeFilter3Cut_L=
			dTimeFilter3Res_L= dTimeFilter3LP_L=
			dTimeFilter3HP_L= dTimeFilter3BP_L=
			dTimeAmp_L= 0.0;

			

		} else {
			dTimeNote1ToNote2_R=
				dTimeOctaveShift_R= dTimeFMPitch_R=
				dTimeFMOsc1_R= dTimeFMOsc2_R=
				dTimeFMOsc3_R= dTimePitchOsc1_R=
				dTimeAmpOsc1_R= dTimePitchOsc2_R=
				dTimeAmpOsc2_R= dTimePitchOsc3_R=
				dTimeAmpOsc3_R= dTimeClip_R=

				dTimeFilter1Cut_R= dTimeFilter1Res_R=
				dTimeFilter1LP_R= dTimeFilter1HP_R=
				dTimeFilter1BP_R= dTimeFilter2Cut_R=
				dTimeFilter2Res_R= dTimeFilter2LP_R=
				dTimeFilter2HP_R= dTimeFilter2BP_R=
				dTimeFilter3Cut_R= dTimeFilter3Res_R=
				dTimeFilter3LP_R= dTimeFilter3HP_R=
				dTimeFilter3BP_R= dTimeAmp_R=
				dTimeTotal = 
				0.0;			
		}
	}

	void ResetTimes()
	{
		ResetTimes(true);
		ResetTimes(false);

	}


	void ResetValues(bool bLeft)
	{
		if(bLeft)
		{
			fValueNote1ToNote2_L= fValueOctaveShift_L=
			fValueFMPitch_L=	fValueFMOsc1_L=
			fValueFMOsc2_L= fValueFMOsc3_L=
			fValuePitchOsc1_L= fValueAmpOsc1_L=
			fValuePitchOsc2_L= fValueAmpOsc2_L=
			fValuePitchOsc3_L= fValueAmpOsc3_L=
			fValueClip_L= fValueFilter1Cut_L=
			fValueFilter1Res_L= fValueFilter1LP_L=
			fValueFilter1HP_L= fValueFilter1BP_L=
			fValueFilter2Cut_L= fValueFilter2Res_L=
			fValueFilter2LP_L= fValueFilter2HP_L=
			fValueFilter2BP_L= fValueFilter3Cut_L=
			fValueFilter3Res_L= fValueFilter3LP_L=
			fValueFilter3HP_L= fValueFilter3BP_L=
			fValueAmp_L= 0.0f;			

		} else {
			fValueNote1ToNote2_R=
				fValueOctaveShift_R= fValueFMPitch_R=
				fValueFMOsc1_R= fValueFMOsc2_R=
				fValueFMOsc3_R= fValuePitchOsc1_R=
				fValueAmpOsc1_R= fValuePitchOsc2_R=
				fValueAmpOsc2_R= fValuePitchOsc3_R=
				fValueAmpOsc3_R= fValueClip_R=

				fValueFilter1Cut_R= fValueFilter1Res_R=
				fValueFilter1LP_R= fValueFilter1HP_R=
				fValueFilter1BP_R= fValueFilter2Cut_R=
				fValueFilter2Res_R= fValueFilter2LP_R=
				fValueFilter2HP_R= fValueFilter2BP_R=
				fValueFilter3Cut_R= fValueFilter3Res_R=
				fValueFilter3LP_R= fValueFilter3HP_R=
				fValueFilter3BP_R= fValueAmp_R=
				0.0f;			
		}
	}
	



	// also need targets for controls

};



// All other parameters are edited through the interface
// except controls

// Set up gvals, tvals and avals
#pragma pack(1)

// storage for global parameters, updated every Tick()
class gvals {
public:
	word amp;
};

// storage for track parameters, updated every Tick()
class tvals {
public:
	
	byte t_Note;
	byte t_Note2;

	word t_SustainLength;
	byte t_LengthUnits;
	byte t_ForceSustainLoop;
	byte t_Smooth;

	word t_ctrl1;
	word t_ctrl2;
	word t_ctrl3;
	word t_ctrl4;

};

#pragma pack()



// Machine Data Input
// for reading from files
// This is a straight mimicry of the Jeskola ones
class CSerpentDataInput
{
private:
	HANDLE hFile;

public:
	void Read(void * pbuf,  int const numbytes)
	{
		DWORD dwBytesRead;
		if(hFile != INVALID_HANDLE_VALUE)
		{
			::ReadFile(hFile, pbuf, numbytes,&dwBytesRead, NULL);

		}
	}

	CSerpentDataInput(TCHAR * lpszFilename)
	{
		hFile = INVALID_HANDLE_VALUE;
		hFile = ::CreateFile(lpszFilename,GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 
			FILE_ATTRIBUTE_NORMAL, NULL);
	}

	virtual ~CSerpentDataInput()
	{
		if(hFile != INVALID_HANDLE_VALUE)
		{
			::CloseHandle(hFile);
		}
	}

	void Read(int &d) { Read(&d, sizeof(int)); }
	void Read(dword &d) { Read(&d, sizeof(dword)); }
	void Read(short &d) { Read(&d, sizeof(short)); }
	void Read(word &d) { Read(&d, sizeof(word)); }
	void Read(char &d) { Read(&d, sizeof(char)); }
	void Read(byte &d) { Read(&d, sizeof(byte)); }
	void Read(double &d) { 	Read(&d, sizeof(double)); }
	void Read(float &d) { 	Read(&d, sizeof(float)); }
	void Read(bool &d) { Read(&d, sizeof(bool)); }

};

// Machine Data Output
// for writing to files
class CSerpentDataOutput
{
private:
	HANDLE hFile;
	
public:
	void Write(void * pbuf,  int const numbytes){
		DWORD dwBytesWritten;
		if(hFile != INVALID_HANDLE_VALUE)
		{
			::WriteFile(hFile, pbuf, numbytes,&dwBytesWritten, NULL);
		}
	}

	CSerpentDataOutput(TCHAR * lpszFilename)
	{
		hFile = INVALID_HANDLE_VALUE;
		hFile = ::CreateFile(lpszFilename,GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 
			FILE_ATTRIBUTE_NORMAL, NULL);
	}

	virtual ~CSerpentDataOutput()
	{
		if(hFile != INVALID_HANDLE_VALUE)
		{
			::CloseHandle(hFile);
		}
	}

	void Write(int d) { Write(&d, sizeof(int)); }
	void Write(dword d) { Write(&d, sizeof(dword)); }
	void Write(short d) { Write(&d, sizeof(short)); }
	void Write(word d) { Write(&d, sizeof(word)); }
	void Write(char d) { Write(&d, sizeof(char)); }
	void Write(byte d) { Write(&d, sizeof(byte)); }
	void Write(double d) { Write(&d, sizeof(double)); }
	void Write(float d) { Write(&d, sizeof(float)); }
	
	void Write(bool d) { Write(&d, sizeof(bool)); }
	void Write(char const *str) { Write((void *)str, int(strlen(str)+1)); }

};




// Meh. All this newfangled stuff - i don't get it.
class miex : public CMDKMachineInterfaceEx {

};



#define MAX_WORK_LENGTH (256)

// dumb lerp. 0 < u < 1
#define LERP(x0,x1,u) ((x0) * (u) + (x1) * (1.0 - (u)))

// smarter - but less accurate - lerp
// precalc increment based on total samples
#define CALC_INCREMENT(x0, x1, totalsamples) (((x1)-(x0)) / (totalsamples))

// then apply increment based on current sample
#define LERP_INCREMENT(x0, inc, sample) ((x0) + ((inc) * (sample)))

#define PI	3.1415926535897932384626433832795
#define MAXAMP 65000

#define ENVELOPES_PER_CHANNEL (29)

class mi : public CMDKMachineInterface {
public:
	mi();
	~mi();

	CMachine * pThisMachine;

	CCriticalSection csEnvelopes;

	// Buzz calls these functions
	void Command(int const i);
	void Tick();
	char const *DescribeValue(int const param, int const value);
	void MDKInit(CMachineDataInput * const pi);
	bool MDKWork(float *psamples, int numsamples, int const mode);
	bool MDKWorkStereo(float *psamples, int numsamples, int const mode);
	void MDKSave(CMachineDataOutput * const po);
	CMDKMachineInterfaceEx *GetEx() { return &ex; }
	void MidiNote(int const channel, int const value, int const velocity);
	void Stop();
	void OutputModeChanged(bool stereo);

	// Serpent load and save
	void SerpentSave(CSerpentDataOutput * const po);
	void SerpentLoad(CSerpentDataInput * const po);

	void SerpentReadEnvelope(CSerpentDataInput * const pi, CSerpentEnvelope * pEnv);
	void SerpentReadPoint(CSerpentDataInput * const pi, CSerpentEnvelopePoint * pPoint );

	void SerpentWriteEnvelope(CSerpentDataOutput * const po, CSerpentEnvelope * pEnv);
	void SerpentWritePoint(CSerpentDataOutput * const po, CSerpentEnvelopePoint * pPoint );



	// File Functions
	void ReadEnvelope(CMachineDataInput * const pi, CSerpentEnvelope * pEnv);
	void ReadPoint(CMachineDataInput * const pi, CSerpentEnvelopePoint * pPoint );

	void WriteEnvelope(CMachineDataOutput * const po, CSerpentEnvelope * pEnv);
	void WritePoint(CMachineDataOutput * const po, CSerpentEnvelopePoint * pPoint );

	// Utility Functions
	int WaveLevels(WaveType wt, int iPhaseInc, float * * pLevel1, float * * pLevel2);

	// Bezier Functions
	serp_double Bezier(serp_double fFirst, serp_double fSecond, serp_double fThird, serp_double fFourth, serp_double u);
	CSerpentEnvelopePoint Bezier(CSerpentEnvelope * env, serp_double u);
	serp_double FindUFromT(CSerpentEnvelope * pStartEnv, serp_double t, CSerpentEnvelope ** ppFoundEnvelope);


	// Extended Machine Interface
	miex ex;

	// Which channel to play with
	bool bLeft;

	// Wavetables
	static int iWaveInitialised;
	static float * ppfSaw[SAMPLE_SETS];					// array of arrays of saw
	static float * ppfSquare[SAMPLE_SETS];				// array of arrays of square
	static float * ppfTriangle[SAMPLE_SETS];			// array of arrays
	static float pfSineSamples[WAVETABLE_SAMPLES];		// array of sine
	static float pfNoise[WAVETABLE_SAMPLES];			// array of noise
	static float pfINoise[WAVETABLE_SAMPLES];			// array of different noise

	

	// Global values
	gvals gval;

	// Track values
	tvals tval[miMAX_TRACKS];
	void SetNumTracks(int const n);

	unsigned int miNumberOfTracks;	// contains the number of tracks in use
	CTrack myTracks[miMAX_TRACKS];

	inline void ResetEnvelopes();

	inline void UpdateEnvelopeArrays();

	inline void GetEnvelopeValues(int numsamples, float * pV1, float * pV2);

	// used
	inline void GetValues(serp_double dStartTime, serp_double * pdEndTime, float * pfStart, float * pfEnd, serp_double dLength, CSerpentEnvelope * pEnv, CTrack * pTrack);

	inline void CalcNote1ToNote2(float *psamples, CTrack * pTrack, int numsamples, bool bLeft);
	inline void CalcOctaveShift(float *psamples, CTrack * pTrack, int numsamples, bool bLeft);
	inline void CalcFMPitch(float *psamples, CTrack * pTrack, int numsamples, bool bLeft);

	inline void CalcFMOut(float *psamples, CTrack * pTrack, int numsamples, bool bLeft);

	inline bool CalcOsc1(float *psamples, CTrack * pTrack, int numsamples, bool bLeft);
	inline bool CalcOsc2(float *psamples, CTrack * pTrack, int numsamples, bool bLeft);
	inline bool CalcOsc3(float *psamples, CTrack * pTrack, int numsamples, bool bLeft);

	inline void CalcClip(float *psamples, CTrack * pTrack, int numsamples, bool bLeft);

	inline void CalcFilter1(float *psamples, CTrack * pTrack, int numsamples, bool bLeft);
	inline void CalcFilter2(float *psamples, CTrack * pTrack, int numsamples, bool bLeft);
	inline void CalcFilter3(float *psamples, CTrack * pTrack, int numsamples, bool bLeft);

	inline bool CalcAmp(float *psamples, CTrack * pTrack, int numsamples, bool bLeft);

	inline bool IsFinished(CTrack * pTrack, bool bLeft);

	// Wave types for the four oscillators
	WaveType wtOsc1;
	WaveType wtOsc2;
	WaveType wtOsc3;
	WaveType wtFM;

	void CalculateWaves();
	void CalculateWavesOriginal();

	static float pfPitch[MAX_WORK_LENGTH];

	static float pfFMPitch[MAX_WORK_LENGTH];
	static float pfFMOut[MAX_WORK_LENGTH];

	// reuse same buffers for each normal osc
	
	static float pfFMOsc[MAX_WORK_LENGTH];
	static float pfPitchOsc[MAX_WORK_LENGTH];
	static float pfAmpOsc[MAX_WORK_LENGTH];

	static float pfOsc1Out[MAX_WORK_LENGTH];
	static float pfOsc2Out[MAX_WORK_LENGTH];
	static float pfOsc3Out[MAX_WORK_LENGTH];

	static float pfClip[MAX_WORK_LENGTH];

	// all steps are done at once for filters
	static float pfFilterLP[MAX_WORK_LENGTH];
	static float pfFilterHP[MAX_WORK_LENGTH];
	static float pfFilterBP[MAX_WORK_LENGTH];

	static float pfOutput[MAX_WORK_LENGTH];

	// and the envelopes.
	// Geez, theres a few of these

	// Left Channel
	CSerpentEnvelope envNote1ToNote2_L;	

	CSerpentEnvelope envOctaveShift_L;

	CSerpentEnvelope envFMPitch_L;
	CSerpentEnvelope envFMOsc1_L;
	CSerpentEnvelope envFMOsc2_L;
	CSerpentEnvelope envFMOsc3_L;

	CSerpentEnvelope envPitchOsc1_L;
	CSerpentEnvelope envAmpOsc1_L;

	CSerpentEnvelope envPitchOsc2_L;
	CSerpentEnvelope envAmpOsc2_L;

	CSerpentEnvelope envPitchOsc3_L;
	CSerpentEnvelope envAmpOsc3_L;

	CSerpentEnvelope envClip_L;

	CSerpentEnvelope envFilter1Cut_L;
	CSerpentEnvelope envFilter1Res_L;
	CSerpentEnvelope envFilter1LP_L;
	CSerpentEnvelope envFilter1HP_L;
	CSerpentEnvelope envFilter1BP_L;

	CSerpentEnvelope envFilter2Cut_L;
	CSerpentEnvelope envFilter2Res_L;
	CSerpentEnvelope envFilter2LP_L;
	CSerpentEnvelope envFilter2HP_L;
	CSerpentEnvelope envFilter2BP_L;

	CSerpentEnvelope envFilter3Cut_L;
	CSerpentEnvelope envFilter3Res_L;
	CSerpentEnvelope envFilter3LP_L;
	CSerpentEnvelope envFilter3HP_L;
	CSerpentEnvelope envFilter3BP_L;

	CSerpentEnvelope envAmp_L;

	// Right Channel
	CSerpentEnvelope envNote1ToNote2_R;

	CSerpentEnvelope envOctaveShift_R;

	CSerpentEnvelope envFMPitch_R;
	CSerpentEnvelope envFMOsc1_R;
	CSerpentEnvelope envFMOsc2_R;
	CSerpentEnvelope envFMOsc3_R;

	CSerpentEnvelope envPitchOsc1_R;
	CSerpentEnvelope envAmpOsc1_R;

	CSerpentEnvelope envPitchOsc2_R;
	CSerpentEnvelope envAmpOsc2_R;

	CSerpentEnvelope envPitchOsc3_R;
	CSerpentEnvelope envAmpOsc3_R;

	CSerpentEnvelope envClip_R;

	CSerpentEnvelope envFilter1Cut_R;
	CSerpentEnvelope envFilter1Res_R;
	CSerpentEnvelope envFilter1LP_R;
	CSerpentEnvelope envFilter1HP_R;
	CSerpentEnvelope envFilter1BP_R;

	CSerpentEnvelope envFilter2Cut_R;
	CSerpentEnvelope envFilter2Res_R;
	CSerpentEnvelope envFilter2LP_R;
	CSerpentEnvelope envFilter2HP_R;
	CSerpentEnvelope envFilter2BP_R;

	CSerpentEnvelope envFilter3Cut_R;
	CSerpentEnvelope envFilter3Res_R;
	CSerpentEnvelope envFilter3LP_R;
	CSerpentEnvelope envFilter3HP_R;
	CSerpentEnvelope envFilter3BP_R;

	CSerpentEnvelope envAmp_R;

	CSerpentEnvelope * envLeftEnvelopes[ENVELOPES_PER_CHANNEL];
	CSerpentEnvelope * envRightEnvelopes[ENVELOPES_PER_CHANNEL];
	CSerpentEnvelope * envAllEnvelopes[ENVELOPES_PER_CHANNEL * 2];


};








//////////////// MFC AppWizard stuff
class CkibibuSerpentApp : public CWinApp
{
public:
	CkibibuSerpentApp();

// Overrides
public:
	virtual BOOL InitInstance();

	DECLARE_MESSAGE_MAP()
};
/////////////// END MFC STUFF
